/*
                                +---------------------------------+
                                |                                 |
                                |       ***   Pyramid   ***       |
                                |                                 |
                                |  Copyright   -tHE SWINe- 2002  |
                                |                                 |
                                |           Pyramid.cpp           |
                                |                                 |
                                +---------------------------------+
*/

#include <malloc.h>
#include <math.h>
#include "Matrix.h"
#include "Load_3ds.h"
#include "Pyramid.h"

enum {
    _Front,
    _Back,
    _Onplane,
    _Split
};
// pozice bodu (face) vi rovin (ped, za, na n, protn)

#define _Len(A) ((float)sqrt(A.x * A.x + A.y * A.y + A.z * A.z))
// spot dlku vektoru

#define _Normalize(A) {float s = _Len(A); A.x /= s; A.y /= s; A.z /= s; }
// vytvo jednotkov vektor

#define _Dot(A,B) (B.x * A.x + A.y * B.y)
// skalrn souin

#define _Cross(A,B,C) C.x = B.y * A.z - A.y * B.z; \
                      C.y = B.z * A.x - A.z * B.x; \
                      C.z = B.x * A.y - A.x * B.y;
// vektorov souin

// spot rovnici roviny ze t bod, lecch na rovin
// (vrcholy nesm leet v jedn pmce)
void SetPlane(Plane *p_plane, float a1, float a2, float a3,
    float b1, float b2, float b3, float c1, float c2, float c3)
{
    Vector u, v, w;

    u.x = c1 - a1;
    u.y = c2 - a2;
    u.z = c3 - a3;
    v.x = b1 - a1;
    v.y = b2 - a2;
    v.z = b3 - a3;
    // 2 vektory, lec v rovin
    
    _Cross(u, v, w);
    _Normalize(w);
    // spot normlov vektor roviny

    p_plane->a = w.x;
    p_plane->b = w.y;
    p_plane->c = w.z;
    p_plane->d = -(p_plane->a * a1 + p_plane->b * a2 + p_plane->c * a3);
    // spot jet komponent D
}

// nastav FOV (field of view - hel vhledu)
// klipovac pyramidy pro rozlien n_Width  n_Height
// a hel, pro kter bylo ureno perspektivn
// zkreslen f_z_delta
void SetFOV(float n_Width, float n_Height, float f_z_delta, Pyramid *p_pyramid)
{
    float x_right, y_bottom, x_left, y_top;

    x_left = -n_Width / (2 * f_z_delta);
    y_top =  n_Height / (2 * f_z_delta);
    x_right =  n_Width / (2 * f_z_delta);
    y_bottom = -n_Height / (2 * f_z_delta);
    // spot souadnice prsek pyramidy s osami X a Y

    SetPlane(&p_pyramid->p0, 0, 0, 0, x_right, 0, 1, x_right, 1, 1);
    SetPlane(&p_pyramid->p1, 0, 0, 0, x_left, 1, 1, x_left, 0, 1);
    SetPlane(&p_pyramid->p2, 0, 0, 0, 0, y_bottom, 1, 1, y_bottom, 1);
    SetPlane(&p_pyramid->p3, 0, 0, 0, 1, y_top, 1, 0, y_top, 1);
    // spot rovnice jednotlivch rovin
}

// ur vzjemnou pozici vertexu a roviny
int inline ClassifyVtxPl(Vertex *p_vertex, Plane *p_plane)
{
    float p;

    p = p_plane->a * p_vertex->x + p_plane->b * p_vertex->y +
        p_plane->c * p_vertex->z + p_plane->d;
    // spot rovnici roviny

    if(p > 0) 
        return _Front;
    if(p < 0)
        return _Back;
    return _Onplane;
}

// vrt 1, pokud vertex le v pyramid
int inline PtInFOV(Vertex *p_vertex, Pyramid *p_pyramid)
{
    return (ClassifyVtxPl(p_vertex, &p_pyramid->p0) != _Back) &
           (ClassifyVtxPl(p_vertex, &p_pyramid->p1) != _Back) &
           (ClassifyVtxPl(p_vertex, &p_pyramid->p2) != _Back) &
           (ClassifyVtxPl(p_vertex, &p_pyramid->p3) != _Back);
}

// ur vzjemnou pozicy face a roviny
int inline ClassifyFacePl(Face *p_face, Plane *p_plane)
{
    int fn = 0, bn = 0, i;

    for(i = 0; i < 3; i ++){
        switch(ClassifyVtxPl(p_face->vertex[i], p_plane)){
        case _Front:
            fn ++;
            break;
        case _Back:
            bn ++;
            break;
        case _Onplane:
            fn ++;
            bn ++;
            break;
        }
    }

    if(fn == 3)
        return _Front;
    if(bn == 3)
        return _Back;
    return _Split;
}

// spot prsek seky, definovan dvma vertexy
// a roviny. pokud oba vrcholy le na sece, vrt
// kopii prvnho vrcholu
Vertex IntersectSegmentPl(Vertex *p_v1, Vertex *p_v2, Plane *p_plane)
{
    float a, b, c, t;
    Vertex vertex;

	a = p_v2->x - p_v1->x;
	b = p_v2->y - p_v1->y;
	c = p_v2->z - p_v1->z;
	// spot vektor z bodu do bodu

    t = p_plane->a * a + p_plane->b * b + p_plane->c * c;
    // spot hel, svran normlou roviny
	// a vektorem z jednoho vertexu do druhho

    if(t == 0)
        return *p_v1;
    // singulrn ppad - oba body le na rovin

    t = (p_v1->x * p_plane->a + p_v1->y * p_plane->b +
	   p_v1->z * p_plane->c + p_plane->d) / t;
    // spot pomr vzdlenost

    vertex.x = p_v1->x - t * (p_v2->x - p_v1->x);
    vertex.y = p_v1->y - t * (p_v2->y - p_v1->y);
    vertex.z = p_v1->z - t * (p_v2->z - p_v1->z);
    vertex.u = p_v1->u - t * (p_v2->u - p_v1->u);
    vertex.v = p_v1->v - t * (p_v2->v - p_v1->v);
    vertex.lu = p_v1->lu - t * (p_v2->lu - p_v1->lu);
    vertex.lv = p_v1->lv - t * (p_v2->lv - p_v1->lv);
    // rozdl souadnice v pomru vzdlenost

    return vertex;
}

// oklipuje face pyramidou. pokud nic nezbude,
// m vracen struktura Polygon nastavenou
// poloku n_vertex_num na 0
Polygon Clip_Face(Face *p_face, Pyramid *p_pyramid)
{
    int a, b, c, d, i;
    Polygon poly;

    a = ClassifyFacePl(p_face, &p_pyramid->p0);
    b = ClassifyFacePl(p_face, &p_pyramid->p1);
    c = ClassifyFacePl(p_face, &p_pyramid->p2);
    d = ClassifyFacePl(p_face, &p_pyramid->p3);
    // otestuje face proti jednotlivm rovinm

    poly.n_vertex_num = 3;
    for(i = 0; i < 3; i ++)
        poly.vertex[i] = p_face->vertex[i][0];
    // zkopruje vertexy do polygonu

    if(a == _Front && b == _Front && c == _Front && d == _Front)
        return poly;
    // pokud je uvnit vech rovin, nic se nedje

    if(a == _Back || b == _Back || c == _Back || d == _Back) {
        poly.n_vertex_num = 0;
        return poly;
    }
    // pokud je za jednou z rovin, nebude vbec vidt

    if(a == _Split)
        poly = SplitPolygon(&poly, &p_pyramid->p0);
    if(b == _Split)
        poly = SplitPolygon(&poly, &p_pyramid->p1);
    if(c == _Split)
        poly = SplitPolygon(&poly, &p_pyramid->p2);
    if(d == _Split)
        poly = SplitPolygon(&poly, &p_pyramid->p3);
    // jinak se mus podle rovin "oezat"

    return poly;
}

// rozdl polygon p_poly rovinou p_plane.
// to co vrac je st, lec ped rovinou
Polygon SplitPolygon(Polygon *p_poly, Plane *p_plane)
{
    int i, prev, status[7];
    Polygon tmp;

    tmp.n_vertex_num = 0;

    for(i = 0; i < p_poly->n_vertex_num; i ++)
        status[i] = ClassifyVtxPl(&p_poly->vertex[i], p_plane);
    // vytvo seznam vech pozic vrchol polygonu

    prev = p_poly->n_vertex_num - 1;
    // pedchoz vrchol

    for(i = 0; i < p_poly->n_vertex_num; i ++) {
        if(status[i] == _Front) {
            if(status[prev] == _Back) {
                tmp.vertex[tmp.n_vertex_num] =
                    IntersectSegmentPl(&p_poly->vertex[prev],
                    &p_poly->vertex[i], p_plane);
                // spot se nov vrchol

                tmp.n_vertex_num ++;
            }
            // hrana prochz plochou

            tmp.vertex[tmp.n_vertex_num] = p_poly->vertex[i];
            tmp.n_vertex_num ++;
            // vrchol je uvnit pyramidy
        } else if(status[prev] == _Front) {
            tmp.vertex[tmp.n_vertex_num] =
                IntersectSegmentPl(&p_poly->vertex[prev],
                &p_poly->vertex[i], p_plane);
            // spot se nov vrchol

            tmp.n_vertex_num ++;
            // hrana prochz plochou
        }

        prev = i;
        // pedchoz vrchol
    }

    //  Yahooo!
    return tmp;
}
